/* zdbs.c
       
   Copyright © 2011, 2012, 2013 Brandon Invergo <brandon@invergo.net>

   This file is part of zeptodb

   zeptodb is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   zeptodb is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with zeptodb.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "zdbs.h"

const char *argp_program_version = "zdbs (zeptodb 1.0)";
const char *argp_program_bug_address = "<brandon@invergo.net>";

static char doc[] =
  "zdbs -- a tool for storing records in a DBM database\v"
  "If an input file is provided, lines are read from it as new records, "
  "otherwise records are added from stdin. Records are expected in the "
  "format KEY|VALUE. If the DATABASE doesn't exist, it will be created.";

static char args_doc[] = "DATABASE";

static struct argp_option options[] = {
  {"input", 'i', "FILE", 0, "Read new records from a file"},
  {"delim", 'd', "CHAR", 0,
   "Character used to separate keys from values (default '|')"},
  {0}
};

struct arguments
{
  char *args[1];
  char *input_file;
  const char *delim;
};

static error_t
parse_opt (int key, char *arg, struct argp_state *state)
{
  struct arguments *arguments = state->input;

  switch (key)
    {
    case 'i':
      arguments->input_file = arg;
      break;

    case 'd':
      if (strlen (arg) > 1)
        argp_usage (state);
      arguments->delim = arg;
      break;

    case ARGP_KEY_ARG:
      if (state->arg_num >= 1)
        argp_usage (state);
      arguments->args[state->arg_num] = arg;
      break;

    case ARGP_KEY_END:
      if (state->arg_num == 0)
        argp_usage (state);
      break;

    default:
      return ARGP_ERR_UNKNOWN;
    }
  return 0;
}

static struct argp argp = { options, parse_opt, args_doc, doc };

int
addrecords (GDBM_FILE database, FILE * input, const char *delim)
{
  size_t linelen = 0;
  char *lineptr = (char *) malloc (linelen);
  char *linecpy;
  char *key_str, *value_str;
  ssize_t chars_read = 0;
  datum key, value;
  int db_status;
  int value_trimmed = 0;

  // add each line of the input to the database
  while (1)
    {
      chars_read = getline (&lineptr, &linelen, input);
      if (chars_read == EOF)
        break;
      if (strrchr(lineptr, '\n'))
        linecpy = strndup (lineptr, chars_read - 1);
      else
        linecpy = strndup (lineptr, chars_read);
      if (!linecpy)
        error (EXIT_FAILURE, errno, "Virtual memory exhausted");
      key_str = strtok (linecpy, delim);
      if (!key_str)
        {
          error (0, errno, "Invalid record format (key%svalue)", delim);
          continue;
        }
      value_str = strtok (NULL, delim);
      if (!value_str)
        {
          error (0, errno, "Invalid record format (key%svalue)", delim);
          continue;
        }
      value.dptr = value_str;
      value.dsize = strlen (value_str);
      key.dptr = key_str;
      key.dsize = strlen (key_str);

      db_status = gdbm_store (database, key, value, GDBM_REPLACE);
      if (db_status != 0)
        error (0, errno, "Could not store record in database");

      free (key.dptr);
    }
  return (0);
}

int
main (int argc, char **argv)
{
  GDBM_FILE database;
  FILE *input;
  struct arguments arguments;
  int result;

  arguments.input_file = "-";
  arguments.delim = "|";

  argp_parse (&argp, argc, argv, 0, 0, &arguments);

  // open the database
  database = gdbm_open (arguments.args[0], 512, GDBM_WRCREAT,
                        S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP, NULL);
  if (!database)
    error (EXIT_FAILURE, errno,
           "Failed to open database %s", arguments.args[0]);

  // if an input file was provided, use it, otherwise use stdin
  if (strcmp (arguments.input_file, "-") == 0)
    input = stdin;
  else
    input = fopen (arguments.input_file, "r");
  if (!input)
    error (EXIT_FAILURE, errno,
           "Failed to open input file %s", arguments.input_file);
  result = addrecords (database, input, arguments.delim);

  fclose (input);
  gdbm_close (database);
  if (result != 0)
    exit (EXIT_FAILURE);
  exit (EXIT_SUCCESS);
}
